#![allow(dead_code)]

mod cuwusions;
mod engine;

use engine::{Component, System};

#[derive(Component)]
struct Sprite {
    sprite: engine::Sprite,
}

#[derive(Component, Default, Clone, Debug)]
struct RigidBody {
    pos: (f64, f64),
    vel: (f64, f64),
    rect: (f64, f64),
    gravity: bool,
}

struct VelocitySystem;
impl System for VelocitySystem {
    fn on_update(&self, ctx: &mut engine::Context, delta: f64) -> Result<(), engine::Error> {
        for id in query!(ctx, RigidBody) {
            let body = ctx.entity_component::<RigidBody>(id);
            // body.pos.0 += body.vel.0 * delta;
            // body.pos.1 += body.vel.1 * delta;
        }
        Ok(())
    }
}

fn rects_collide(
    rect: (f64, f64),
    pos: (f64, f64),
    other_rect: (f64, f64),
    other_pos: (f64, f64),
) -> bool {
    pos.0 + rect.0 > other_pos.0
        && pos.0 < other_pos.0 + other_rect.0
        && pos.1 + rect.1 > other_pos.1
        && pos.1 < other_pos.1 + other_rect.1
}

fn rec_point_collides(rect: (f64, f64), pos: (f64, f64), point: (f64, f64)) -> bool {
    pos.0 + rect.0 < point.0 && pos.0 > point.0 && pos.1 + rect.1 < point.1 && pos.1 > point.1
}

#[derive(Component, Clone, Default)]
struct Collider {
    resolve: bool,
    colliding_surface: Option<Surface>,
}

fn horizontal_line_intersect(p0: (f64, f64), vel: (f64, f64), line_y: f64) -> Option<(f64, f64)> {
    if vel.1 == 0.0 {
        // exclusively going left or right, ergo there's no collision with a horizontal line
        return None;
    } else if vel.0 == 0.0 {
        // no change in x, ergo the intersect must be at p0_x
        return Some((p0.0, line_y));
    }
    // y = ax + b
    let a = vel.1 / vel.0;
    let b = p0.1 - p0.0 * a;

    let x = (line_y - b) / a;
    Some((x, line_y))
}

fn vertical_line_intersect(p0: (f64, f64), vel: (f64, f64), line_x: f64) -> Option<(f64, f64)> {
    if vel.0 == 0.0 {
        // exclusively going up or down, ergo there's no collision with a vertical line
        return None;
    } else if vel.1 == 0.0 {
        // no change in y, ergo the intersect must be at p0_y
        return Some((line_x, p0.1));
    }
    //
    // y = ax + b
    let a = vel.1 / vel.0;
    let b = p0.1 - p0.0 * a;

    let y = a * line_x + b;
    Some((line_x, y))
}

#[derive(Clone, Debug)]
enum Surface {
    Top,
    Bottom,
    Left,
    Right,
}

fn point_distance(a: (f64, f64), b: (f64, f64)) -> f64 {
    ((a.0 - b.0).abs().powi(2) + (a.1 - b.1).abs().powi(2)).sqrt()
}

fn point_rect_closest_surface(
    p0: (f64, f64),
    vel: (f64, f64),
    rect_pos: (f64, f64),
    rect: (f64, f64),
) -> Option<(f64, Surface)> {
    let closest_horizontal_intersect = [
        (horizontal_line_intersect(p0, vel, rect_pos.1), Surface::Top),
        (
            horizontal_line_intersect(p0, vel, rect_pos.1 + rect.1),
            Surface::Bottom,
        ),
    ]
    .into_iter()
    .filter_map(|(intersect, surface)| intersect.map(|intersect| (intersect, surface)))
    .map(|(point, surface)| (point_distance(p0, point), surface))
    .min_by(|(dist_a, _), (dist_b, _)| dist_a.total_cmp(dist_b));

    let closest_vertical_intersect = [
        (vertical_line_intersect(p0, vel, rect_pos.0), Surface::Left),
        (
            vertical_line_intersect(p0, vel, rect_pos.0 + rect.0),
            Surface::Right,
        ),
    ]
    .into_iter()
    .filter_map(|(intersect, surface)| intersect.map(|intersect| (intersect, surface)))
    .map(|(point, surface)| (point_distance(p0, point), surface))
    .min_by(|(dist_a, _), (dist_b, _)| dist_a.total_cmp(dist_b));

    match (closest_horizontal_intersect, closest_vertical_intersect) {
        (Some(left), Some(right)) => {
            if right.0 < left.0 {
                Some(right)
            } else {
                Some(left)
            }
        }
        (None, v) | (v, None) => v,
    }
}

struct CollisionSystem;
impl System for CollisionSystem {
    fn on_update(&self, ctx: &mut engine::Context, delta: f64) -> Result<(), engine::Error> {
        for id in query!(ctx, RigidBody, Collider) {
            let collider = ctx.entity_component::<Collider>(id);
            collider.colliding_surface = None;
            let collider = ctx.entity_component::<Collider>(id).clone();
            if !collider.resolve {
                continue;
            }
            let body = ctx.entity_component::<RigidBody>(id).clone();
            for other_id in query!(ctx, RigidBody, Collider) {
                if id == other_id {
                    continue;
                }
                let other_body = ctx.entity_component::<RigidBody>(other_id).clone();
                let new_pos = cuwusions::rects_collision_resolution_pos(
                    body.pos,
                    (body.vel.0 * delta, body.vel.1 * delta),
                    body.rect,
                    other_body.pos,
                    other_body.rect,
                );

                let body = ctx.entity_component::<RigidBody>(id);
                body.pos = new_pos;
            }
        }
        Ok(())
    }
}
struct GravitySystem;
impl System for GravitySystem {
    fn on_update(&self, ctx: &mut engine::Context, delta: f64) -> Result<(), engine::Error> {
        for id in query!(ctx, RigidBody) {
            let body = ctx.entity_component::<RigidBody>(id);
            if !body.gravity {
                continue;
            }
            body.vel.1 = if body.vel.1 < 800.0 {
                body.vel.1 + 1600.0 * delta
            } else {
                body.vel.1
            };
        }
        Ok(())
    }
}

#[derive(Component)]
struct Cloud;

struct CloudSystem;
impl System for CloudSystem {
    fn on_update(&self, ctx: &mut engine::Context, delta: f64) -> Result<(), engine::Error> {
        let cloud_amount = ctx.entities_with_component::<Cloud>().len();
        if cloud_amount < 1 {
            let cloud = ctx.load_sprite("textures/clouds.png").unwrap();
            spawn!(
                ctx,
                Cloud,
                Sprite { sprite: cloud },
                RigidBody {
                    pos: (-100.0, 150.0),
                    ..Default::default()
                },
            );
        }

        for id in query!(ctx, Cloud, RigidBody) {
            let body = ctx.entity_component::<RigidBody>(id);
            body.vel.0 = if body.vel.0 < 200.0 {
                body.vel.0 + 200.0 * delta
            } else {
                body.vel.0
            };
        }

        for id in query!(ctx, Cloud, RigidBody) {
            let body = ctx.entity_component::<RigidBody>(id);
            if body.pos.0 > 1400.0 {
                ctx.despawn(id);
            }
        }
        Ok(())
    }
}

struct SpriteRenderer;
impl System for SpriteRenderer {
    fn on_update(&self, ctx: &mut engine::Context, _delta: f64) -> Result<(), engine::Error> {
        for id in query!(ctx, Sprite, RigidBody) {
            let body = ctx.entity_component::<RigidBody>(id).clone();
            let sprite = ctx.entity_component::<Sprite>(id).sprite;

            ctx.draw_sprite(sprite, body.pos.0 as i32, body.pos.1 as i32)?;
        }
        Ok(())
    }
}

#[derive(Component)]
struct PlayerMovement;

struct PlayerMovementSystem;
impl System for PlayerMovementSystem {
    fn on_update(&self, ctx: &mut engine::Context, _delta: f64) -> Result<(), engine::Error> {
        for id in query!(ctx, PlayerMovement, RigidBody, Collider) {
            let d_down = ctx.key_pressed(engine::Keycode::D);
            let a_down = ctx.key_pressed(engine::Keycode::A);
            let w_down = ctx.key_pressed(engine::Keycode::W);
            let collider = ctx.entity_component::<Collider>(id).clone();
            let body = ctx.entity_component::<RigidBody>(id);
            body.vel.0 = if d_down && !a_down {
                400.0
            } else if !d_down && a_down {
                -400.0
            } else {
                0.0
            };
            if let (Some(Surface::Top), true) = (collider.colliding_surface, w_down) {
                body.vel.1 = -1000.0;
            }
        }
        Ok(())
    }
}

fn main() {
    let mut game = engine::Game::new().unwrap();

    let mut context = game.context();
    context.add_system(VelocitySystem);
    context.add_system(CollisionSystem);
    context.add_system(SpriteRenderer);
    context.add_system(PlayerMovementSystem);
    context.add_system(GravitySystem);
    context.add_system(CloudSystem);
    let player = context.load_sprite("textures/player.png").unwrap();
    let background = context.load_sprite("textures/literally_dprk.png").unwrap();
    let nope = context.load_sprite("textures/nuh-uh.png").unwrap();

    spawn!(
        &mut context,
        Sprite { sprite: background },
        RigidBody::default(),
    );

    spawn!(
        &mut context,
        Sprite { sprite: player },
        RigidBody {
            pos: (400.0, 400.0),
            rect: (128.0, 128.0),
            gravity: true,
            ..Default::default()
        },
        Collider {
            resolve: true,
            ..Default::default()
        },
        PlayerMovement,
    );

    spawn!(
        &mut context,
        RigidBody {
            pos: (184.0, 540.0),
            rect: (960.0, 128.0),
            ..Default::default()
        },
        Collider::default(),
    );

    spawn!(
        &mut context,
        RigidBody {
            pos: (300.0, 200.0),
            rect: (32.0, 32.0),
            ..Default::default()
        },
        Collider::default(),
        Sprite { sprite: nope },
    );

    spawn!(
        &mut context,
        RigidBody {
            pos: (900.0, 400.0),
            rect: (32.0, 32.0),
            ..Default::default()
        },
        Collider::default(),
        Sprite { sprite: nope },
    );

    game.run();
}
